import { defineComponent, PropType, h, ConcreteComponent, onMounted, watch, computed, isRef, unref, Component, getCurrentInstance } from 'vue'
import { useRoute, RouteLocationNormalizedLoaded } from 'vue-router'
import { Meta, SchemaNodeList, NormalizedSchemaNode, Data, ChildSchemaNode } from '../schema'
import { pluginDriver } from '../plugins'
import { useLoading, OpenLoading, CloseLoading } from './hooks/useLoading'
import useHook from './hooks/useHook'
import userParser from './hooks/userParser'
import { isFalse, isString, normalizeSchema } from '../util'

export type RendererProps = NormalizedSchemaNode

export type ParserUtil = {
    openLoading: OpenLoading
    closeLoading: CloseLoading
    route: RouteLocationNormalizedLoaded
}

export const SchemaRenderer: Component<RendererProps> = defineComponent({
    name: 'SchemaRenderer',
    props: {
        uniqueKey: {
            type: String,
            required: true,
        },
        type: {
            type: String,
            required: true,
        },
        label: {
            type: String,
        },
        meta: {
            type: Object as PropType<Meta>,
            default: () => ({})
        },
        items: {
            type: Array as PropType<SchemaNodeList>
        }
    },
    setup: (props, context) => {
        const route = useRoute()
        const {
            closeLoading,
            openLoading,
        } = useLoading()
        const parserUtil: ParserUtil = {
            closeLoading,
            openLoading,
            route,
        }

        const {component, props: parsedProps, data: componentData, events: componentEvents, isInit, children, methods} = userParser(props as NormalizedSchemaNode, context, parserUtil)
        const componentProps = computed( () => {
            if(!parsedProps) return undefined

            const props: Data = isRef(parsedProps) ? unref(parsedProps) : parsedProps

            return props && Object.entries(props).reduce((result, [key, value]) => Object.assign(result, {
                [key]: isRef(value) ? unref(value) : value
            }), {})
        })

        // 调用解析器
        const { callHook } = useHook(props as NormalizedSchemaNode)

        onMounted( () => {
            callHook('mounted')

            isFalse(isInit?.value) && openLoading()
        })

        watch( () => isInit?.value, (newV) => {
            newV && closeLoading()
        })

        // created
        callHook('created')

        return {
            isInit,
            component,
            componentProps,
            componentData,
            componentEvents,
            children,
            closeLoading,
            openLoading,
            ...methods,
        }
    },
    render(){
        const {component, componentProps, componentEvents, isInit, children, uniqueKey, $slots: {
                default: defaultSlotsOfRenderer
            }
        } = this
        const initialSlots = {
            default: []
        }

        let slots: any = children?.reduce((currentSlots: Data, item: ChildSchemaNode | string) => {
            const { slot } = item as ChildSchemaNode
            const childComponent = isString(item) ? item : h(SchemaRenderer as any, {
                ...normalizeSchema(item),
                key: `${uniqueKey}-${item.key}`
            })

            if(slot){
                currentSlots[slot] = () => childComponent
            }else{
                currentSlots.default.push(childComponent)
            }

            return currentSlots
        }, initialSlots)

        if(defaultSlotsOfRenderer){
            if(!slots) slots = initialSlots

            slots.default.push(defaultSlotsOfRenderer())
        }
        // SchemaRenderer支持slots,
        // || slotsOfRenderer

        // 将default重置为函数形式
        if(slots){

            const {default: defaultSlot } = slots

            // if()

            defaultSlot.length && (slots.default = () => defaultSlot)
        }

        return h((component as ConcreteComponent<RendererProps>), {
            ...componentProps,
            ...componentEvents,
            ref: 'ui',
        } as RendererProps, slots)
    }
})