import { EvaluatorTypes } from '../constant'
import logger from '../util/logger'
import { BaseEvaluator, EvaluatorOptions } from './base'
import PathEvaluator from './path'

export {
    PathEvaluator,
}

export * from './base'

export type EvaluatorFactoryType = {
    _createdEvaluators: Map<string, BaseEvaluator>
    create(evaluatorType: EvaluatorTypes): BaseEvaluator
    _getEvaluator(evaluatorType: EvaluatorTypes): BaseEvaluator | undefined
    _setEvaluator(evaluatorType: EvaluatorTypes, evaluator: BaseEvaluator): void
    _createEvaluator(evaluatorType: EvaluatorTypes): BaseEvaluator
}

export const EvaluatorFactory: EvaluatorFactoryType = {
    _createdEvaluators: new Map(),
    create(type){
        let evaluator = this._getEvaluator(type)

        if(!evaluator){
            evaluator = this._createEvaluator(type)
            this._setEvaluator(type, evaluator)
        }

        return evaluator
    },
    _getEvaluator(type){
        return this._createdEvaluators.get(type)
    },
    _setEvaluator(type, evaluator){
        this._createdEvaluators.set(type, evaluator)
    },
    _createEvaluator(type){
        switch (type) {
            case EvaluatorTypes.Path:
                return new PathEvaluator()
            case EvaluatorTypes.Expression:
                return new PathEvaluator()
        }

        logger.error(`unknown Evaluator Type: ${type}`)
    }
}

export const Evaluator = {
    eval(type: EvaluatorTypes, target: unknown, options: EvaluatorOptions): any{
        const evaluator = EvaluatorFactory.create(type)

        evaluator.setOptions(options)

        const reslut = evaluator.eval(target)

        // console.info(`${evaluator.constructor?.name} eval`, target, 'to', reslut, `with options: `, options)

        return reslut
    },
    /**
     * 解析带有表达式的字符串
     *
     * @param   {String}             string        输入的字符串
     * @param   {EvaluatorOptions}   options       Evaluator配置
     *
     * @return  {String}                           转换后的字符串
     */
    evalString(string: string, options: EvaluatorOptions){
        const hasVariable = string.indexOf('{') > -1

        if (!hasVariable) return string

        return string.replace(/\$\{(.+)\}/g, (matched, expression) => this.eval(EvaluatorTypes.Path, expression, options))
    }
}