// 解析器管理
import { SimpleManager } from "./common/simple";
import { BasePlugin, Util } from '../plugins'
import { PluginTypes, ActionTypes, OtherMetaTypes, AllMetaTypes } from '../constant'
import { Data, Meta, RestArguments, ActionMeta } from '../schema'
import logger from "../util/logger";
import { isConstantProp } from "../util";

export interface BaseAction {
    type: string
}

export type Action<T = Data> = T & BaseAction
export type ActionHandler = (action: Action, util?: Util, ...args: any[]) => any

export interface ActionPlugin extends BasePlugin {
    type: PluginTypes.Action,
    specificMeta: string, // 该类型Action所特有的meta字段，可以根据该字段唯一确定该类型Action
    dependentMetas: string[], // 该类型依赖的所有meta字段，包括specificMeta
    handler: ActionHandler
}

export type PluginSet = Set<ActionPlugin>

export type MetaRecords = {
    [key in AllMetaTypes]?: Meta
}

const setRecord = (records: Data, type: AllMetaTypes, key: string, value: any) => {
    let record = records[type]

    if(!record){
        record = {}
        records[type] = record
    }

    record[key] = value
}

const shouldSetRecord = (type: string | undefined, currentType: AllMetaTypes) => !type || type === currentType

export class ActionManager extends SimpleManager{
    // 内部缓存对象
    protected dependentMetaPluginMap: Map<string, PluginSet> // 依赖的Meta到对应ActionPlugin的映射
    protected metaPluginMap: Map<string, ActionPlugin> // ActionPlugin依赖的Meta Key到Plugin的映射
    constructor(){
        super()
        this.name = 'Action'
        this.carReset = false
        this.dependentMetaPluginMap = new Map()
        this.metaPluginMap = new Map()
    }

    /**
     * 判断一个Meta是否为一个Action Meta
     *
     * @param   {Meta}     meta  待判定的Action Meta
     *
     * @return  {boolean}        结果
     */
    isActionMeta(meta: Meta): boolean{
        return !!(meta && Object.keys(meta).some( key => this.metaPluginMap.has(key) ))
    }

    /**
     * 判断meta是否为某个特定类型的Action Meta
     *
     * @param   {string}  type  Action的类型
     * @param   {Meta}    meta  待判定的Action Meta
     *
     * @return  {boolean}       结果
     */
    isActionMetaByType(type: string, meta: Meta): boolean{
        const action = this.get(type) as ActionPlugin

        return !!(meta && Object.keys(meta).some( key => key === action.specificMeta ))
    }

    handleAction(meta: ActionMeta, util: Util, restArgs?: RestArguments): any{
        const action = this.getAction(meta)

        if(!action) return logger.warn(`can not get acticon type of meta: `, meta)

        return this.callActionPlugin(action, {
            ...meta,
            type: action.name
        }, util, restArgs)
    }

    callActionPlugin(action: ActionPlugin, meta: Action<ActionMeta>, util: Util, restArgs: RestArguments = []){
        const { handler } = action

        return handler.call(null, meta, util, ...restArgs)
    }

    handleActionByType(type: string, meta: ActionMeta, util: Util, restArgs?: RestArguments){
        const action = this.get(type) as ActionPlugin

        return this.callActionPlugin(action, {
            ...meta,
            type
        }, util, restArgs)
    }

    /**
     * 将混合在一起的Meta分解成一组一组的特定类型的Meta
     * 如果指定type，则只提取该类型的Meta，其他的全部放到other里
     *
     * @param   {Meta}          meta  待分解的Meta
     * @param   {AllMetaTypes}  type  是否指定特定类型
     *
     * @return  {MetaRecords}         key为Meta类型，值为对应Meta的集合
     */
     destructureMeta(meta: Meta = {}, type?: AllMetaTypes): MetaRecords{
        return Object.entries(meta as Data).reduce((records, [key, value]) => {
            // 处理:key这种直接传给组件的情况
            if(isConstantProp(key) && shouldSetRecord(type,  OtherMetaTypes.Props)){
                setRecord(records, OtherMetaTypes.Props, key, value)

                return records
            }

            // key属于ActionMeta里的字段
            const plugin = this.getPluginBySpecificMeta(key)
            const actionType = plugin?.name as ActionTypes

            if(plugin && shouldSetRecord(type, actionType)){
                plugin.dependentMetas.forEach( key => key in meta && setRecord(records, actionType, key, meta[key]))

                return records
            }

            const dependedPlugins = this.getPluginByDependentMetaKey(key)

            if(dependedPlugins){
                const willBeSetByDependedPlugin = [...dependedPlugins].some( plugin => shouldSetRecord(type, plugin.name as ActionTypes) )

                if(willBeSetByDependedPlugin) return records
            }

            // 都不是，就是other的情况
            setRecord(records, OtherMetaTypes.Other, key, value)

            return records
        }, {} as MetaRecords)
    }

    getPluginBySpecificMeta(key: string): ActionPlugin | undefined{
        return this.metaPluginMap.get(key)
    }

    getPluginByDependentMetaKey(key: string): PluginSet | undefined{
        return this.dependentMetaPluginMap.get(key)
    }

    getAction(meta: ActionMeta): ActionPlugin | undefined{
        let action

        for (const key in meta) {
            if (Object.prototype.hasOwnProperty.call(meta, key)) {
                action = this.getPluginBySpecificMeta(key)

                if(action) break
            }
        }

        return action
    }

    // 设置specificMeta和
    setPluginMap(item: ActionPlugin){
        const {metaPluginMap, dependentMetaPluginMap} = this

        metaPluginMap.set(item.specificMeta, item)

        item.dependentMetas.forEach( key  => {
            let pluginSet = dependentMetaPluginMap.get(key)

            if(!pluginSet){
                pluginSet = new Set([item])
                dependentMetaPluginMap.set(key, pluginSet)
            }

            pluginSet.add(item)
        })
    }

    add(key: string, item: ActionPlugin): any{
        const isSuccess = super.add(key, item)

        if(isSuccess){
            this.setPluginMap(item)
        }
    }
}