import { SetupContext, ref, computed, reactive } from 'vue'
import { ComponentNode, Meta, SchemaNode, Data } from '../../schema'
import { ParserPlugin } from '../../manager/parser'
import { Util } from '../../plugins'
import { ParserTypes, PluginTypes, UITypes, MODEL_VALUE, UPDATE_MODEL_HANDLER } from '../../constant'
import useSource from './useSource'
import useVisible from './useVisible'
import useTitle from '../hooks/useTitle'

export type DialogMethods = {
    setSource(source: Data): void
    open(): void,
    close(): void
}

enum ConfigType{
    Props = 'props',
    Events = 'events'
}

type EventsRecord = Record<string, Meta>

type ConfigRecord = {
    props: Meta,
    events: EventsRecord
}

type SpecialChild = ParserTypes.Form

const SPECIAL_CHILDREN: Array<SpecialChild> = [ParserTypes.Form]

type ConfigRecordMap = Record<SpecialChild, ConfigRecord>



// 弹窗子组件默认配置
const DIALOG_CHILDREN_DEFAULT_PROPS_MAP: ConfigRecordMap = {
    Form: {
        props: {
            columnsCount: 1,
            // actionsShow: false
        },
        events: {
            confirm: {
                type: 'submit',
                form: '$specialKey'
            }
        }
    }
}

const getDialogChildrenDefaultConfig = (childType: SpecialChild, type = ConfigType.Props, eventType?: string): Meta | EventsRecord | undefined => {
    const childConfig = DIALOG_CHILDREN_DEFAULT_PROPS_MAP[childType]

    if(!childConfig) return undefined

    const defaultConfig = childConfig[type]

    if(type === ConfigType.Events && eventType){
        return (defaultConfig as EventsRecord)[eventType]
    }

    return defaultConfig
}

export default {
    name: ParserTypes.Dialog,  // 插件名称
    type: PluginTypes.Parser,
    ui: UITypes.Dialog,
    parse: (schema: SchemaNode, context: SetupContext, {
        openLoading,
        closeLoading,
        pluginDriver
    }: Util) => {
        const { visible, setVisible, open, close } = useVisible()
        const { source, setSource } = useSource()
        const {title} = useTitle(schema, pluginDriver, {
            alias: {
                '.': 'source'
            },
            source: {
                source
            }

        })
        const children = computed( () => schema.items?.map( ({type, meta, ...other}) => ({ type, ...other, meta: Object.assign({source}, getDialogChildrenDefaultConfig(type as SpecialChild), meta) })) )
        const specialChild = computed( () => schema.items?.find(item => SPECIAL_CHILDREN.includes(item.type as SpecialChild)) )
        const onConfirm = () => {
            // 没有需要特殊操作的子组件，直接关闭
            if(!specialChild.value) return close()

            openLoading()

            const {type, key: specialKey} = specialChild.value
            const actionConfig = getDialogChildrenDefaultConfig(type as SpecialChild, ConfigType.Events, 'confirm')
            const action = pluginDriver.evalOptionsWithVariableKey(actionConfig, {
                source: {
                    specialKey
                }
            })

            Promise.resolve(pluginDriver.handleAction(action)).then( () => {
                closeLoading()
                close()
            })
        }
        const props = computed(() => ({
            ...schema.meta,
            title,
            [MODEL_VALUE]: visible,
            [UPDATE_MODEL_HANDLER]: setVisible,
            onConfirm,
            onCancel: close,
            footerShow: false,
        }))

        return {
            data: {
                source,
            },
            props,
            methods: {
                open,
                close,
                setSource,
            } as DialogMethods,
            children
        } as ComponentNode
    },
} as ParserPlugin