import { Ref, ref } from 'vue'
import { SchemaRenderer } from '../../components/renderer'
import { AppConfig } from '../../engine'
import { useRouter, useRoute, RouteRecordRaw } from 'vue-router'
import { ParserTypes } from '../../constant'
import { PluginDriver } from '../../plugins'
import useInit from '../hooks/useInit'

type MenuItem = {
    index?: boolean // 是否为首
    label: string, // 菜单文案
    key: string, // 路由路径及name
    subs?: Menus,
    subPages?: Menus,
}

// kcp menus接口返回的数据
type MenusResponse = {
    appTitle: string
    menus: Array<MenuItem>
    subPages: Array<MenusResponse>
}

type RouteRecords = Array<RouteRecordRaw>

type Menus = Array<MenuItem>

const menuItem2RouteRecord = ({key, label: title, index}: MenuItem, isFirstLevel?: boolean): RouteRecordRaw => ({
    path: isFirstLevel ? `/${key}` : key,
    name: key,
    meta: {
        title,
        index,
    },
    component: SchemaRenderer,
    props: {
        uniqueKey: key,
        type: ParserTypes.Page
    },
})
const subPages2Children = (subPages: Menus): RouteRecords => subPages.map<RouteRecordRaw>( ({subPages, ...subPage}) => ({
    ...menuItem2RouteRecord(subPage),
    children: (subPages && subPages.length) ? subPages2Children(subPages) : undefined
}))

// 将接口返回的菜单配置，转换为routes
const collectPages = (menus: Menus, collectedPages: RouteRecords = []): RouteRecords => menus?.reduce<RouteRecords>((collectedPages, menu) => {
    const {subs, subPages} = menu
    if(subs && subs.length) return collectPages(subs, collectedPages)

    collectedPages.push({
        ...menuItem2RouteRecord(menu, true),
        children: (subPages && subPages.length) ? subPages2Children(subPages) : undefined
    })

    return collectedPages
}, collectedPages)

export default function useMenus(appConfig: AppConfig, pluginDriver: PluginDriver){
    const menus: Ref<Menus> = ref([])
    const { isInit, initRequest } = useInit(false, pluginDriver)
    const title = ref(appConfig?.title)
    const { kcpApi, menuApiPath, appKey } = appConfig
    const router = useRouter()
    const currentRoute = useRoute()
    const getRoutes = async () => {
        const res: MenusResponse = await initRequest({
            url: kcpApi + menuApiPath,
            method: 'get',
            params: {
                app_key: appKey
            }
        })

        const { appTitle, menus: menusInResponse } = res
        const newRoutes = collectPages(menusInResponse) || []

        newRoutes.forEach((route: RouteRecordRaw) => router.addRoute(route));

        title.value = appTitle
        menus.value = menusInResponse
    }

    getRoutes().then( () => {
        // 路由正确解析，直接返回
        if(currentRoute.name) return

        const routes = router.getRoutes()
        const indexRoute = routes.find( route => route.meta.index )
        const routeOfCurrentPath = router.resolve(currentRoute.path)

        if(routeOfCurrentPath.name){
            router.push(routeOfCurrentPath)
        }else if(indexRoute){
            router.push(indexRoute)
        }
    })

    return {
        isInit,
        title,
        currentRoute,
        menus,
        getRoutes
    }
}