import { Component, SetupContext, ComponentPublicInstance } from 'vue'
import { SchemaNode, ComponentNode, Meta, NormalizedSchemaNode, Data, ActionMeta } from '../schema'
import BASE_PLUGINS from './base'
import { ParserManager, ParserPlugin, ActionManager, ActionPlugin, CustomPluginManager, CustomPlugin, FormatterManager, FormatterPlugin, FormatType } from '../manager'
import { UIManager, UIConfig } from '../manager/ui'
import { Config, AppConfig } from '../engine'
import { ParserUtil } from '../components/renderer'
import { isUndefined, http, HttpRequestConfig } from '../util'
import logger from '../util/logger'
import { ParserTypes, PluginTypes, UITypes, AllMetaTypes, OtherMetaTypes, ActionTypes, EvaluatorTypes } from '../constant'
import { Plugin as PluginUI } from 'vue-seed-ui'
import { ComponentManager } from '../manager/component'
import { HTTP_ERROR_MESSAGE } from "../constant/http";
import { Evaluator, EvaluatorOptions } from '../evaluators'

export interface BasePlugin {
    name: string,  // 插件名称
    type?: string, // 插件类型，默认为UI类型
}

export type HttpUtil = (options: HttpRequestConfig) => Promise<any>

// 解析组件辅助工具集
export interface PluginUtil {
    http: HttpUtil
    appConfig: AppConfig | null
    pluginDriver: PluginDriver
    Message?: PluginUI
}

export type Util = PluginUtil & ParserUtil

export type ParseResult = ComponentNode | Promise<ComponentNode>

export type Plugin = ParserPlugin | ActionPlugin | CustomPlugin<any>
type Plugins = Plugin[]
export type UserPlugins = Plugins | undefined

export type ActionMetaRecords = {
    [key in ActionTypes]?: Meta
}

export type ActionOption = {
    action: Meta
    source: Data
    type: string,
    alias: Data,
}

export class PluginDriver{
    options: Config | null
    plugins: Plugins
    parserManager: ParserManager
    actionManager: ActionManager
    uiManager: UIManager
    componentManager : ComponentManager
    customPluginManager: CustomPluginManager
    formatterManager: FormatterManager
    // 辅助工具集合
    util: PluginUtil | null
    message?: PluginUI
    private static instance: PluginDriver

    constructor(){
        this.options = null
        this.util = null
        this.plugins = []

        // init managers
        this.uiManager = new UIManager()
        this.parserManager = new ParserManager()
        this.actionManager = new ActionManager()
        this.componentManager = new ComponentManager()
        this.customPluginManager = new CustomPluginManager()
        this.formatterManager = new FormatterManager()

        this.init()
    }

    init(){
        this.distributePlugins(BASE_PLUGINS)
        this.initUtil()
    }

    initUtil(){
        this.util = {
            http: (options: HttpRequestConfig) => http(options).catch( err => {
                this.message?.error(err?.message || HTTP_ERROR_MESSAGE)

                return Promise.reject(err)
            }),
            appConfig: null,
            pluginDriver: this
        }
    }

    setUtil(key: keyof PluginUtil, value: any){
        (this.util as Util)[key] = value
    }

    setOptions(options: Config){
        const {
            ui,
            plugins,
            appConfig
        } = options
        const {util} = this

        this.options = options

        plugins && this.distributePlugins(plugins)
        this.addUI(ui)
        util && (util.appConfig = appConfig)
    }

    // 根据插件类型，对插件进行分发
    distributePlugins(plugins: Plugins){
        const { plugins: installedPlugins } = this
        plugins.forEach( plugin => {
            const {type, name} = plugin

            if(installedPlugins.includes(plugin)) return logger.warn(`${type} plugin ${name} has installed, please don't install it repeatedly`, plugin)

            installedPlugins.push(plugin)

            switch(type){
                case PluginTypes.Parser:
                    this.parserManager?.add(name, plugin)
                    break
                case PluginTypes.Action:
                    this.actionManager?.add(name, plugin as ActionPlugin)
                    break
                case PluginTypes.Formatter:
                    this.formatterManager?.add(name, plugin as CustomPlugin<any>)
                    break
                case PluginTypes.Plugin:
                    this.customPluginManager?.add(name, plugin as FormatterPlugin)
                    break
                default:
                    logger.warn(`unknown type of plugin ${name} with type ${type}`, plugin)
            }
        })
    }

    runHook(pluginName: string, hookName: string, context?: object, ...args: any[]){
        return this.customPluginManager.getPlugins(pluginName)?.reduce( (prevValue, plugin) => {
            const hook = plugin[hookName as keyof Plugin]

            if(!hook) return prevValue

            return typeof hook === 'function' ?  hook.apply(context, args) : hook
        }, undefined) || Promise.resolve(args && args[0])
    }

    /**
     * 获取SchemaNode对应的解析器Parser
     *
     * @param   {SchemaNode}  options  [options description]
     *
     * @return  {[type]}               [return description]
     */
    getParser(options: NormalizedSchemaNode): ParserPlugin{
        const {type} = options as SchemaNode
        const parserManager = this.parserManager as ParserManager
        const parser = parserManager.get(type) || parserManager.get(ParserTypes.Common)

        return parser as ParserPlugin
    }

    formatValue(formatInfo: any, input: any, type: FormatType = FormatType.ToDisplay){
        return this.formatterManager.format(formatInfo, input, type)
    }

    getAppConfig(): AppConfig | null{
        const { options } = this

        if(options === null) return null

        const { appConfig } = options

        return appConfig
    }

    addComponent(key: string, componentInstance?: ComponentPublicInstance | null){
        componentInstance && this.componentManager.add(key, componentInstance)
    }

    clearComponents(){
        this.componentManager.clear()
    }

    getUIInstance(key: string): ComponentPublicInstance | undefined{
        return this.componentManager.getUIInstance(key)
    }

    getComponent(key: string): ComponentPublicInstance | undefined{
        return this.componentManager.get(key) as ComponentPublicInstance
    }

    parseSchema(options: NormalizedSchemaNode, context: SetupContext, util: ParserUtil): ComponentNode{
        const {message: Message} = this
        const parserPlugin = this.getParser(options)
        const { parse, ui } = parserPlugin
        const pluginUtil: Util = Object.assign({
            Message
        }, this.util, util)
        const componentNode = parse.call(null, options, context, pluginUtil)
        let { component } = componentNode

        if(isUndefined(component)){
            componentNode.component = (typeof ui === 'string' ? this.getUI(ui) : ui) as Component
        }

        return componentNode
    }

    handleActionWithVariable({action, source, type, alias}: ActionOption, ...restArgs: any[]){
        if(source) action = this.evalOptionsWithVariableKey(action, {source, alias})

        type = action?.type || type

        if(type) return this.handleActionByType(type, action, ...restArgs)

        return this.handleAction(action, ...restArgs)
    }

    handleAction(meta: Meta, ...restArgs: any[]): any{
        const {actionManager, util} = this

        if(!actionManager || !actionManager.isActionMeta(meta)) return logger.warn(`unknown action type with meta: `, meta)

        return actionManager.handleAction(meta as ActionMeta, util as Util, restArgs)
    }

    handleActionByType(type: string, meta: Meta, ...restArgs: any[]): any{
        const {actionManager, util} = this

        if(!actionManager || !actionManager.isActionMetaByType(type, meta)) return logger.warn(`unknown action type: ${type} with meta: `, meta)

        return actionManager.handleActionByType(type, meta as ActionMeta, util as Util, restArgs)
    }

    destructureMeta(meta: Meta, type?: AllMetaTypes){
        return this.actionManager.destructureMeta(meta, type)
    }

    evalOptionsWithVariableKey(target: unknown, options: EvaluatorOptions): any{
        return Evaluator.eval(EvaluatorTypes.Path, target, options)
    }

    getUI(type: string): Component{
        return this.uiManager?.get(type) as Component
    }

    addUI(ui: UIConfig){
        const {uiManager} = this

        uiManager?.addCaches(ui)

        // 设置完Message之后，要重新设置util的http
        if(UITypes.Message in ui){
            this.message = uiManager?.get(UITypes.Message) as PluginUI
        }
    }

    static getInstance(){
        return this.instance ? this.instance : (this.instance = new PluginDriver())
    }
}

export const pluginDriver = PluginDriver.getInstance()