import { Data, SchemaNode, NormalizedSchemaNode, NormalizedSchema } from '../schema'
export * from './http'

// 类型转换

/**
 * 将字符串转为驼峰形式
 */
export const toCamelCase = (str: string): string => str?.replace(/[-_](\w)/g, (matched, letter) => letter.toUpperCase())

/**
 * 将驼峰形式字符串转为大驼峰(帕斯卡)形式
 */
export const toPascalCase = (str: string): string => toCamelCase(str)?.replace(/^\w/, letter => letter.toUpperCase())

// 类型判断相关

const toString = Object.prototype.toString

export const toRawType = (input: any): string => toString.call(input).slice(8, -1)

export const isObject = (val: unknown): val is Record<any, any> => val !== null && typeof val === 'object'

export const isPlainObject = (val: unknown): val is object => toString.call(val) === '[object Object]'

export const isArray = (val: unknown): val is Array<any> => Array.isArray(val)

export const isFunction = (val: unknown): val is Function => typeof val === 'function'

export const isPromise = <T = any>(val: unknown): val is Promise<T> => {
    return isObject(val) && isFunction(val.then) && isFunction(val.catch)
}

export const isUndefined = (val: unknown): val is undefined => val === undefined

type Defined = boolean | string | number | any[] | Data

export const isDefined = (val: unknown): val is Defined => val !== undefined && val !== null

export const isFalse = (val: unknown): val is false => val === false

export const isString = (val: unknown): val is string => typeof val === 'string'

export const isEmptyObject = (val: object): boolean => !!(val && Object.keys(val).length)

export const mergeObjects = (target: Data, source: unknown): Data => {
    if(!isObject(source)) return target

    Object.entries(source).forEach(([key, value]) => {
        target[key] = value
    })

    return target
}

export const normalizeSchema = (schema: SchemaNode): NormalizedSchemaNode => {
    const normalizedSchema: NormalizedSchemaNode = Object.assign({
        uniqueKey: schema.key
    } as NormalizedSchema, schema)

    return normalizedSchema
}

// 是否为可以直接传给组件的属性，以:开头，例如:type="primary"
export const isConstantProp = (prop: string) => prop.startsWith(':')

export const getConstantProp = (prop: string) => prop.replace(/^\:/, '')

export const NOOP = (...args: any[]): any => {}

// 确保对象的指定路径有值，没有时将其设置为默认值
export const ensureValueByPath = (source: Data, path: string, defaultValue: any = {}): void => {
    const pathArr = path.split('.')
    const lastKey = pathArr.pop() as string

    while (pathArr.length) {
        const key = pathArr.shift() as string

        if (!(key in source)) {
            source[key] = {}
        }

        source = source[key]
    }

    if (!(lastKey in source)) source[lastKey] = defaultValue
}

function getArrayIndex(str: string): number | null {
    const arrayIndexReg = /^\[([\d]+)\]$/
    const result = str.match(arrayIndexReg)

    return result && Number(result[1])
}

/**
 * 根据path获取对象的值，支持.的方式取对象属性, [i]的方式取数组的元素
 *
 * @param  {String} path   取值路径，对象属性以.分隔，数组以[i]取值，例如 a[1].b 表示target的属性a的第二个元素的属性b的值
 * @param  {Data}   source 取值的源数据
 * @return {any}           取得的值
 */
export const getValueByPath = (path: string, source?: Data | null): any => {
    if(!source) return ''

    const splitReg = /\[?([\w\-_$]+)\]?/g
    const isArrIndex = /^\[[\w\-_$]+\]$/
    const pathArr: string[] = path && path.match(splitReg) || []
    let value = source
    let key: string

    while (value && pathArr && pathArr.length) {
        key = pathArr.shift() as string

        if (!key) continue

        // [1]这样的表示数组的路径的key
        if (isArrIndex.test(key)) {
            const index = getArrayIndex(key)

            if (!index && index !== 0) {
                throw new Error(`can not get index of sub path ${key} in path ${path}`)
            }
            if (!isArray(value)) {
                throw new Error(`target ${JSON.stringify(value)} is not Array, can not get array like sub path ${key} in path ${path}`)
            }

            value = value[index]
        } else if (isPlainObject(value)) {
            value = (value as Data)[key]

            if(!isDefined(value)) (value as string) = ''
        } else {
            (value as unknown as string) = ''
        }
    }

    return value
}